use serde_yaml;
use serde::{Deserialize, Serialize};
use octocrab;
use clap::Parser;
use std::future::Future;
use std::time::Duration;
use futures;
use tokio;
use log::*;
use chrono::prelude::*;
use frankenstein;
use frankenstein::client_reqwest::Bot;
use frankenstein::AsyncTelegramApi;
use frankenstein::SendMessageParams;

#[derive(Debug, PartialEq, Serialize, Deserialize)]
struct Repo {
    owner: String,
    repo: String,
    check: String,
}

impl Repo {
    async fn init(&mut self) {
        tokio::time::sleep(Duration::from_millis(0)).await;
        let _ = create_repo_files(&self.owner, &self.repo).await;
    }
    async fn resolve(&mut self) {
        let opts: Opts = Opts::parse();
        let random_spred = rand::random::<u64>();

        println!("{} - LOG - {}/{} will be checked at {} ({}s interval + {}s random spread)", 
            Local::now().trunc_subsecs(0).to_rfc3339(), &self.owner, &self.repo, 
            hhmmss(opts.interval+(random_spred % opts.spread)).unwrap(), 
            opts.interval, random_spred % opts.spread);
        
        tokio::time::sleep(Duration::from_secs(opts.interval+(random_spred % opts.spread))).await;
        match &self.check[..] {
            "tag" => {
                let _ = github_get_latest_tag(&self.owner, &self.repo).await;
            }
            _ => {
                let _ = github_get_latest_release(&self.owner, &self.repo).await;
            }
        };
    }

}

#[derive(Parser, Debug)]
#[clap(author, version, about, long_about = None)]
struct Opts {
    /// Config file
    #[clap(long, default_value = "config.yaml")]
    config: String,
    /// Data dir
    #[clap(long, default_value = "data")]
    datadir: String,
    /// Telegram token
    #[clap(long)]
    token: String,
    /// Check interval
    #[clap(long, default_value = "28800")]
    interval: u64,
    /// Spread interval
    #[clap(long, default_value = "600")]
    spread: u64,
    /// Telegram chat id
    #[clap(long, allow_hyphen_values(true))]
    chatid: i64,
    /// Verbose level, accept multiple occurrences
    #[clap(short, long, action = clap::ArgAction::Count)]
    verbose: u8,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let opts: Opts = Opts::parse();
    stderrlog::new()
        .module(module_path!())
        .verbosity(usize::from(opts.verbose+1))
        .timestamp(stderrlog::Timestamp::Second)
        .quiet(false)
        .init()
        .unwrap();
    
    println!("{} - LOG - Config file name: {}", Local::now().trunc_subsecs(0).to_rfc3339(), opts.config);
    debug!("Using token: {}", opts.token);
    println!("{} - LOG - Using chat id: {}", Local::now().trunc_subsecs(0).to_rfc3339(), opts.chatid);
    println!("{} - LOG - Using check interval {}s + random spread from {}", 
        Local::now().trunc_subsecs(0).to_rfc3339(), opts.interval, opts.spread);

    let config: Vec<Repo> = get_config(&opts.config).await?;
    let items: Vec<_> = config.iter()
        .map(|x| Repo { owner: x.owner.trim().to_string(), repo: x.repo.trim().to_string(), check: x.check.trim().to_string() })
        .collect();

    let _items = join_parallel(items.into_iter().map(|mut item| async {
        item.init().await;
        item
    })).await;

    loop {
        let items: Vec<_> = config.iter()
            .map(|x| Repo { owner: x.owner.trim().to_string(), repo: x.repo.trim().to_string(), check: x.check.trim().to_string() })
            .collect();
        let _items = join_parallel(items.into_iter().map(|mut item| async {
            item.resolve().await;
            item
        })).await;
    }
}

async fn join_parallel<T: Send + 'static>(
    futs: impl IntoIterator<Item = impl Future<Output = T> + Send + 'static>,
) -> Vec<T> {
    let tasks: Vec<_> = futs.into_iter().map(tokio::spawn).collect();
    futures::future::join_all(tasks)
        .await
        .into_iter()
        .map(Result::unwrap)
        .collect()
}

async fn get_config(filename: impl AsRef<std::path::Path>) -> Result<Vec<Repo>, Box<dyn std::error::Error>> {
    let config = std::fs::File::open(filename)?;
    let contents: Vec<Repo> = serde_yaml::from_reader(config)?;
    Ok(contents)
}

async fn create_repo_files(owner: &String, repo: &String) -> Result<(), Box<dyn std::error::Error>> {
    debug!("Create repo file {}-{}", owner, repo);
    let opts: Opts = Opts::parse();
    let _repofile = tokio::fs::OpenOptions::new().write(true).create(true).open(opts.datadir.to_string() 
        + "/" + &owner.trim().to_string() + "-" + &repo.trim().to_string()).await?;
    Ok(())
}

async fn read_file(filename: impl AsRef<std::path::Path>) -> Result<String, Box<dyn std::error::Error>>{
    let contents = tokio::fs::read_to_string(filename).await?;
    Ok(contents)
}

async fn overwrite_repo_file(filename: impl AsRef<std::path::Path>, release_tag: &String) -> Result<(), Box<dyn std::error::Error>> {
    let mut file = tokio::fs::File::create(filename).await?;
    tokio::io::AsyncWriteExt::write_all(&mut file, release_tag.as_bytes()).await?;
    Ok(())
}


async fn telegram_notify(_release_tag: &String, release_url: &String) -> Result<(), Box<dyn std::error::Error>> {
    let opts: Opts = Opts::parse();
    let api = Bot::new(opts.token.as_str());
    let send_message_params: SendMessageParams = SendMessageParams::builder()
        .chat_id(opts.chatid)
        .text(release_url)
        .build();

    if let Err(err) = api.send_message(&send_message_params).await {
        println!("Failed to send message: {err:?}");
    }

    Ok(())
}

async fn github_get_latest_release(owner: &String, repo: &String) -> octocrab::Result<(), Box<dyn std::error::Error>> {
    let opts: Opts = Opts::parse();
    let release = octocrab::instance()
        .repos(owner, repo)
        .releases()
        .get_latest()
        .await?;
    let version_in_file = read_file(opts.datadir.to_string() + "/" + &owner.trim().to_string() + "-" + &repo.trim().to_string()).await?;

    match version_in_file.to_string() == release.tag_name.to_string() {
        true => {
            info!("{}/{}: current known version is {:?}", owner, repo, version_in_file.to_string());
            println!("{} - LOG - {}/{}: known version matches with latest release {}", 
                Local::now().trunc_subsecs(0).to_rfc3339(), owner, repo, release.tag_name.to_string());
        }
        _ => {
            info!("{}/{}: current known version is {:?}", owner, repo, version_in_file.to_string());
            println!("{} - LOG - {}/{}: known version does not match with latest release {}. Sending new release URL {}", 
                Local::now().trunc_subsecs(0).to_rfc3339(), owner, repo, release.tag_name.to_string(), release.html_url);
            
            let _ = overwrite_repo_file(opts.datadir.to_string() + "/" + &owner.to_string() 
                + "-" + &repo.to_string(), &release.tag_name.to_string()).await?; 

            let _ = telegram_notify(&release.tag_name.to_string(), &release.html_url.to_string()).await?;
        }
    };
    Ok(())
}

async fn github_get_latest_tag(owner: &String, repo: &String) -> octocrab::Result<(), Box<dyn std::error::Error>> {
    let opts: Opts = Opts::parse();
    let tags = octocrab::instance()
        .repos(owner, repo)
        .list_tags()
        .send()
        .await?;

    let github_url = "https://github.com/".to_string() + 
        &owner.trim().to_string() + "/" + 
        &repo.trim().to_string()  + "/releases/tag/" + 
        &tags.items[0].name.to_string();

    let version_in_file = read_file(opts.datadir.to_string() + "/" + &owner.trim().to_string() + "-" + &repo.trim().to_string()).await?;

    match version_in_file.to_string() == tags.items[0].name.to_string() {
        true => {
            info!("{}/{}: current known version is {:?}", owner, repo, version_in_file.to_string());
            println!("{} - LOG - {}/{}: known version matches with latest tag {}", 
                Local::now().trunc_subsecs(0).to_rfc3339(), owner, repo, tags.items[0].name.to_string());
        }
        _ => {
            info!("{}/{}: current known version is {:?}", owner, repo, version_in_file.to_string());
            println!("{} - LOG - {}/{}: known version does not match with latest tag {}. Sending new tag URL {}", 
                Local::now().trunc_subsecs(0).to_rfc3339(), owner, repo, tags.items[0].name.to_string(), &github_url);
            
            let _ = overwrite_repo_file(opts.datadir.to_string() + "/" + &owner.to_string() 
                + "-" + &repo.to_string(), &tags.items[0].name.to_string()).await?; 

            
            let _ = telegram_notify(&tags.items[0].name.to_string(), &github_url).await?;
        }
    };
    Ok(())
}

fn hhmmss(seconds: u64) -> Result<String, Box<dyn std::error::Error>> {
    Ok((Local::now() + chrono::Duration::seconds(seconds as i64)).trunc_subsecs(0).to_rfc3339())
}

